#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

class FileExistsError(Exception):
    """FileExistsError
    exists to avoid problems with python2
    """
    pass

class LTGException(Exception):
    """LTGException"""
    pass

class MissingDirectoryException(LTGException):
    """MissingDirectoryException"""
    pass

class NoPrefixException(LTGException):
    """NoPrefixException"""
    pass

class NonRecursiveException(LTGException):
    """NonRecursiveException"""
    pass

class LinkException(LTGException):
    """LinkException"""
    pass

class NoOwnedFileException(LTGException):
    """NoOwnedFileException"""
    pass

class NoStoredFileException(LTGException):
    """NoStoredFileException"""
    pass

class NoCategoryException(LTGException):
    """NoCategoryException"""
    pass

class GitCommandException(LTGException):
    """GitCommandException"""
    pass

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import os, logging

from . import file_utils, storage, linker

log = logging.getLogger(__name__)

class Manager():
    """Manager

    :storage_dir: the directory where files should be stored
    :linker_dir: the directory where stored files should be linked to
    """
    def __init__(self, storage_dir, linker_dir):
        self._storage = storage.Storage(storage_dir)
        self._linker = linker.Linker(linker_dir)

    def git(self, *args):
        """git

        :param *args: arguments to git
        :returns: the git command standard output
        """
        return self._storage.git(*args)

    def storage_dir(self):
        """storage_dir"""
        return self._storage.root()

    def linker_dir(self):
        """linker_dir"""
        return self._linker.root()

    def get_categories(self):
        """get_categories

        :returns: absolute paths to all categories
        """
        for directory in os.listdir(self.storage_dir()):
            if not directory.startswith('.'):
                yield os.path.join(self.storage_dir(), directory)

    def get_category_files(self, category):
        """get_category_files

        :param category: the requested category
        :returns: generator object for all files in categories
        """
        for root,dirs,files in os.walk(category):
            dirs[:] = [ d for d in dirs if d != '.git' ]

            for d in dirs:
                yield os.path.join(root, d)

            for f in files:
                yield os.path.join(root, f)

            # TODO break?
            break

    def owns(self, path):
        """owns

        TODO

        :path: a path to an existing file.
        :returns: boolean if path belongs to FileHandler.

        """
        return (self._storage.owns(path) and self._linker.owns(path))

    def store(self, paths, relative_to, category, force):
        """store

        TODO

        :path: a path to an existing file.

        """
        if relative_to is None:
            relative_to = self.linker_dir()

        base_dir = os.path.join(self.storage_dir(), category)

        # TODO committing
        for path in paths:
            new_path = file_utils.get_target_path(base_dir,
                                                  os.path.abspath(path),
                                                  relative_to)

            self._storage.add(path, new_path, force)

    def link(self, force):
        """link

        TODO

        :param force: if existing files should be overwritten

        """
        for category in self.get_categories():
            for f in self.get_category_files(category):
                link_path = file_utils.get_target_path(self.linker_dir(),
                                                       f,
                                                       category)

                self._linker.add(f, link_path, force)

    def move(self, path, new_path, force):
        """TODO: Docstring for move.

        :path: a path to an existing (owned) file.
        :new_path: the new path.
        :returns: the new path.

        """
        # TODO think about this
        try:
            return self._linker.move(path, new_path, force)
        except exceptions.NoOwnedFileException:
            return self._storage.move(path, new_path, force)

    def remove(self, path, recursive, force):
        """TODO: Docstring for remove.

        :path: a path to an existing file.

        """
        raise NotImplementedError("remove")

        self._linker.remove(path, recursive)

        storage_path = None # TODO

        self._storage.remove(storage_path, recursive)

    def unlink(self, files):
        """unlink

        :param path:
        """
        raise NotImplementedError("unlink")

        self._linker.remove(path)

        stored_file = self._storage.get_target_path(path, self.linker_dir)

        file_utils.copy(stored_file, path)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import os, logging

from abc import ABCMeta, abstractmethod

from . import utils, file_utils, exceptions

class FileHandler(object):
    """FileHandler

    :param root:
    """

    def __init__(self, root):
        self._logger = logging.getLogger(utils.fullname(self))
        self._root_dir = root
        file_utils.makedir(self._root_dir)

    def check(self):
        """check"""
        self._logger.debug('checking')

        if not os.path.exists(self._root_dir):
            raise exceptions.MissingDirectoryException(self._root_dir)
        elif not os.path.isdir(self._root_dir):
            raise exceptions.FileExistsError(self._root_dir)

    def root(self):
        """root"""
        return self._root_dir

    def owns(self, path):
        """owns_path

        :param path:
        """
        self._logger.debug('checking if "%s" is owned path', path)

        return file_utils.is_prefix(self._root_dir, path)

    def add(self, path, new_path, force):
        """add

        :param path:
        :param new_path:
        :param force:
        """
        self._logger.debug('adding %s to %s', path, new_path)

        if not self.owns(new_path):
            raise exceptions.NoOwnedFileException(new_path)

        self._add(path, new_path, force)

        return new_path

    def move(self, path, new_path, force):
        """TODO: Docstring for move.

        :param path: a path to an existing (owned) file.
        :param new_path: the new path.
        :returns: the new path.

        """
        if not self.owns(path):
            raise exceptions.NoOwnedFileException(path)
        elif not self.owns(new_path):
            raise exceptions.NoOwnedFileException(new_path)

        self._move(path, new_path, force)

        return new_path

    def remove(self, path, recursive):
        """remove

        :param path:
        :param recursive:
        """
        if not self.owns(path):
            raise exceptions.NoOwnedFileException(path)

        self._remove(path, recursive)

        return path

    @abstractmethod
    def _add(self, path, new_path, force):
        """TODO: Docstring for link.

        :param path: a path to an existing file.
        :param new_path: a path to an existing file.
        :returns: the path of the resulting (owned) file.
        """
        raise NotImplementedError("_add in %s" % self.__class__.__name__)

    @abstractmethod
    def _move(self, path, new_path, force):
        """_move

        :param path:
        :param new_path:
        :param force:
        """
        raise NotImplementedError("_move in %s" % self.__class__.__name__)

    @abstractmethod
    def _remove(self, path, recursive):
        """_remove

        :param path:
        :param recursive:
        """
        raise NotImplementedError("_remove in %s" % self.__class__.__name__)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import os, click, logging, coloredlogs

from . import manager, utils, exceptions, version

log = logging.getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(short_help=__name__, context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug/--no-debug', default=False)
@click.option('-v', '--verbose/--no-verbose', default=False)
@click.option('-s', '--storage-dir', type=click.Path(),
        help='where to store files')
@click.option('-l', '--link-dir', type=click.Path(),
        help='where to link stored files')
@click.pass_context
def cli(ctx, debug, verbose, storage_dir, link_dir):
    """
    Store files in a git repository and replace them with links while keeping
    the original structure.
    """
    if debug:
        loglevel = logging.DEBUG
    elif verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING

    coloredlogs.install(level=loglevel, show_timestamps=False,
                                        show_hostname=False,
                                        show_name=debug)

    if storage_dir is None:
        storage_dir = os.path.expanduser('~/.ltg')
    if link_dir is None:
        link_dir = os.path.expanduser('~/')

    ctx.obj['manager'] = manager.Manager(storage_dir, link_dir)

# store

@cli.command(short_help='add a new files to storage')
@click.argument('files', type=click.Path(exists=True), nargs=-1, required=False)
@click.option('--relative-to', type=click.Path())
@click.option('-c', '--category', default='general')
@click.option('-f', '--force/--no-force', default=False)
@click.pass_context
def store(ctx, relative_to, category, force, files):
    """
    Store a new file in the storage directory, leave the original.
    """
    ctx.obj['manager'].store(files, relative_to, category, force)

# link

@cli.command(short_help='link stored files')
@click.option('-f', '--force/--no-force', default=False)
@click.pass_context
def link(ctx, force):
    """
    Link all stored files to the linker directory.
    """
    ctx.obj['manager'].link(force)

# add (store + link)

@cli.command(short_help='add and link new files')
@click.argument('files', type=click.Path(exists=True), nargs=-1, required=False)
@click.option('--relative-to', type=click.Path())
@click.option('-c', '--category', default='general')
@click.option('-f', '--force/--no-force', default=False)
@click.pass_context
def add(ctx, files, relative_to, category, force):
    """
    Store a new file in the storage directory and replace the original with a
    link.
    """
    ctx.obj['manager'].store(files, relative_to, category, force)
    ctx.obj['manager'].link(True)

# unlink

@cli.command(short_help='unlink files or categories')
@click.option('-c', '--category')
@click.option('-r', '--recursive/--non-recursive', default=False)
@click.argument('files', type=click.Path(), nargs=-1)
@click.pass_context
def unlink(ctx, files, category, recursive):
    """
    Replace linked files with their original.
    """
    ctx.obj['manager'].unlink(files)

# rm

@cli.command(short_help='remove stored files or categories')
@click.option('--keep-original/--remove-original', default=True)
@click.option('--keep-link/--remove-link', default=False)
@click.option('-r', '--recursive/--no-recursive', default=False)
@click.option('-f', '--force/--no-force', default=False)
@click.argument('files', type=click.Path(), nargs=-1)
@click.pass_context
def rm(ctx, keep_original, keep_link, recursive, force, files):
    """
    Remove files from the storage directory.
    """
    ctx.obj['manager'].remove(files, recursive, force)

# mv

@cli.command(short_help='move stored files')
@click.option('-f', '--force/--no-force', default=False)
@click.argument('src', type=click.Path(exists=True))
@click.argument('dst', type=click.Path())
@click.pass_context
def mv(ctx, force, src, dst):
    """
    Move files from one place to the other.
    """
    ctx.obj['manager'].move(src, dst, force)

# git

@cli.command(short_help='run git commands in storage dir')
@click.argument('args', nargs=-1)
@click.pass_context
def git(ctx, args):
    """
    Send commands to git, use '--' to avoid ltg interpreting commands as its own:

    $ ltg git -- commit --interactive

    """
    ctx.obj['manager'].git(*args)

# info

@cli.command(short_help='print general information')
@click.pass_context
def info(ctx):
    """
    Print version and used directories.
    """
    print('ltg %s' % version.__version__)
    print('storage: %s' % ctx.obj['manager'].storage_dir())
    print('linker: %s' % ctx.obj['manager'].linker_dir())

###############

def exception_msg(exception):
    """exception_msg

    :param exception:
    """
    return "%s: %s" % (exception.__class__.__name__, exception)

def run():
    """run"""
    try:
        cli(obj={})
    except exceptions.FileExistsError as e:
        log.error("file exists, use --force to ovewrite: %s", e)
        exit(1)
    except (NotImplementedError, exceptions.LTGException) as e:
        log.error(exception_msg(e))
        exit(1)
    except Exception as e:
        log.exception(e)
        exit(1)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import logging

from . import core, file_utils, exceptions

class Linker(core.FileHandler):
    """Linker"""

    def _add(self, path, new_path, force):
        """_add

        :param path:
        :param new_path:
        :param force:
        """
        file_utils.link(path, new_path, force)

    def _move(self, path, new_path, force):
        """move

        :param path:
        :param new_path:
        :param force:
        """
        file_utils.move(path, new_path, force)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import logging

from . import exceptions, file_utils

log = logging.getLogger(__name__)

def user_confirms(prompt):
    """user_confirms

    :param prompt: the message prompt
    :returns: True if user input is empty or equal to 'y', 'Y' or 'yes'
    """
    return input(prompt + ' [y]/n ') in ['','y','Y','yes']

def ensure_directory(path, interactive=False):
    """ensure_directory

    :param path: the supposed directory
    :param interactive: if user interaction should be required to create the directory
    :returns: True if path is a directory, else False
    """
    if not os.path.exists(path):
        if interactive and not user_confirms('%s does not exist, want to create it?' % path):
            return False

        file_utils.makedir(path)
    return True

def fullname(o):
    """fullname

    :param o: an object
    :returns: the fully qualified object name
    """
    return o.__module__ + "." + o.__class__.__name__

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

from . import cli

cli.run()

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8



# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import os, shutil, logging

from . import exceptions

log = logging.getLogger(__name__)

def get_target_path(base, path, relative_to):
    """
    TODO: Docstring for get_target_path.
    TODO: better name

    >>> get_target_path('/abc/def', '/home/lala/abc/1234', '/home')
    '/abc/def/lala/abc/1234'

    >>> get_target_path('/abc/def', '/abc/def/lala/123', '/home')
    Traceback (most recent call last):
    ...
    NoPrefixException: ('/home', '/abc/def/lala/123')

    >>> get_target_path('/abc/def', '/home/lala/123', '/var')
    Traceback (most recent call last):
    ...
    NoPrefixException: ('/var', '/home/lala/123')

    :param path:
    :param relative_to:

    """
    if not is_prefix(relative_to, path):
        raise exceptions.NoPrefixException(relative_to, path)

    return os.path.join(base, path.replace(relative_to, '').lstrip('/'))

def filetype(path):
    """filetype

    :param path:
    """
    if os.path.exists(path):
        if os.path.isdir(path):
            return 'directory'
        elif os.path.islink(path):
            return 'link'
        elif os.path.isfile(path):
            return 'file'
        else:
            return 'unkown'
    elif os.path.islink(path):
        return 'broken link'

def is_prefix(prefix, path):
    """is_prefix

    >>> is_prefix('/', '/')
    True

    >>> is_prefix('/var', '/foo/bar')
    False

    >>> is_prefix('/', '/foo/bar')
    True

    >>> is_prefix('/var', '/var/abc')
    True

    :param prefix:
    :param path:
    """
    return (os.path.commonprefix([prefix, path]) == prefix)

def is_broken_symlink(path):
    """is_broken_symlink

    :param path:
    """
    return not os.path.exists(os.readlink(path))

def link_file(target, path, force):
    """link

    :target:
    :path:
    :force:
    """
    if os.path.exists(path):
        if os.path.samefile(target, path):
            return

        elif not force:
            raise exceptions.FileExistsError(path)
        else:
            os.remove(path)

    makedir(os.path.dirname(path))

    log.debug('linking file %s to %s', target, path)

    os.symlink(target, path)

def link_directory(target, path, force):
    """_link_directory

    :param target:
    :param path:
    :param force:
    """
    log.debug("linking directory %s to %s", target, path)

    for root,dirs,files in os.walk(target):
        dirs[:] = [ d for d in dirs if d != '.git' ]

        new_root = get_target_path(path, root, target)

        makedir(new_root)

        for f in files:
            link_file(os.path.join(root, f), os.path.join(new_root, f), force)

def link(target, path, force):
    """link

    :param target:
    :param path:
    :param force:
    """
    if os.path.isdir(target):
        link_directory(target, path, force)
    else:
        link_file(target, path, force)

def remove(path):
    """remove.

    :path:
    """
    if os.path.exists(path) or os.path.islink(path):
        log.debug('removing %s in %s', filetype(path), path)

        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)

def move(src, dst, force):
    """move

    :param src:
    :param dst:
    :param force:
    """
    if os.path.exists(dst):
        if not force:
            raise exceptions.FileExistsError(dst)
        else:
            os.remove(dst)

    log.debug('moving %s from %s to %s', filetype(src), src, dst)

    shutil.move(src, dst)

def copy(src, dst, force):
    """copy

    :param src:
    :param dst:
    :param force:
    """
    if os.path.exists(dst):
        if not force:
            raise exceptions.FileExistsError(dst)
        else:
            os.remove(dst)

    log.debug('copying %s from %s to %s', filetype(src), src, dst)

    makedir(os.path.dirname(dst))

    if os.path.isdir(src):
        shutil.copytree(src, dst)
    else:
        shutil.copy(src, dst)

def makedir(path, force = False):
    """makedir

    :param path:
    :param force:
    """
    if os.path.exists(path):
        if os.path.isdir(path):
            return

        elif not force:
            raise exceptions.FileExistsError(path)
        else:
            os.remove(path)

    log.debug('creating directory %s' % path)

    os.makedirs(path)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8




import os, pip

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

package = 'ltg'

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def read_version():
    version = {}
    with open('%s/version.py' % package, 'rb') as init_file:
        exec(init_file.read(), version)

    return version['__version__']

def read_requirements():
    install_reqs = pip.req.parse_requirements('requirements.txt', session=pip.download.PipSession())
    return [str(ir.req) for ir in install_reqs if ir is not None]

setup(name             = package,
      version          = read_version(),
      author           = 'Aljosha Friemann',
      author_email     = 'aljosha.friemann@gmail.com',
      description      = 'link to git',
      license          = None,
      keywords         = ["utility", "configuration", "git", "linux"],
      url              = 'https://bitbucket.org/afriemann/ltg',
      download_url     = 'https://pypi.python.org/pypi/ltg',
      packages         = [package],
      install_requires = read_requirements(),
      long_description = read('README'),
      classifiers      = [],
      entry_points     = {'console_scripts': ['ltg=%s.cli:run' % package]}
)



# -*- coding: utf-8 -*-
"""
.. module:: TODO
   :platform: Unix
   :synopsis: TODO.

.. moduleauthor:: Aljosha Friemann <aljosha.friemann@gmail.com>

"""

import logging

from sh import git, ErrorReturnCode_128
logging.getLogger('sh').setLevel(logging.WARNING)

from . import core, file_utils, exceptions

class Storage(core.FileHandler):
    """Storage

    :param storage_dir: the directory where files should be stored
    """

    def __init__(self, storage_dir):
        super(Storage, self).__init__(storage_dir)
        self._git = git.bake('--git-dir=%s/.git' % storage_dir,
                             '--work-tree=%s' % storage_dir)
        try:
            self.git('status')
        except ErrorReturnCode_128:
            self.git('init')

    def _add(self, path, new_path, force):
        """_add

        :param path: to path to be stored
        :param new_path: the path where the file should be stored
        :param force: if existing files should be overwritten
        """
        file_utils.copy(path, new_path, force)
        self._git('add', new_path)

    def _move(self, path, new_path, force):
        """_move

        :param path: to original file path
        :param new_path: the new path
        :param force: if existing files should be overwritten
        """
        self._git('mv', path, new_path)

    def git(self, *args):
        """git

        :param *args:
        """
        self._logger.debug("git command: %s" % list(args))
        return self._git(*args)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4 fenc=utf-8
